function [A,B,T] = coupled_tensors(X,Y,R,varargin)
%COUPLED_TENSORS computes the non-negative coupled factorization of
%tensors X and Y.
%   X is decomposed as sum_r a_r o b_r o t_r, while 
%   Y is decomposed as sum_r a_r o a_r o t_r .
%
%   [A,B,T] = COUPLED_TENSORS(X,Y,R) computes an estimate of the
%   best rank-R PARAFAC factorization of coupled tensors X and Y,
%   as described above, with nonnegativity constraints on the
%   factors.
%   This version uses multiplicative updates, first introduced by
%   Lee & Seung in their NMF algorithm. 
%
%   [A,B,T] = COUPLED_TENSORS(X,Y,R,'param',value,...) specifies
%   optional parameters and values. 
%   Valid parameters and their default values are:
%        'alpha' - Weight of Y's reconstruction error. {(normX/normY)^2}
%        'maxiters' - Maximum number of iterations. {10}


%% Default values and initializing constants
normX = norm(X);
normY = norm(Y);

alpha = (normX/normY)^2;
maxiters = 10;

%% Parsing parameters
if (rem(length(varargin),2) == 1)
  error('Optional parameters must go in pairs.');
else 
  for i=1:2:(length(varargin)-1)
    if ~ischar(varargin{i})
      error('Unknown type of optional parameters name (parameter names must be strings).');
    end

    switch lower(varargin{i})
      case 'alpha'
	alpha = varargin{i+1};
      case 'maxiters'
        maxiters = varargin{i+1};
      otherwise 
	warning(['Unrecognized parameter: ''' varargin{i} '''']);
    end;
  end;
end;


%% Extract number of dimensions and norms.
N = size(X, 1);
M = size(X, 2);
K = size(X, 3);

%% Set up for iterations - initializing A,B,T and the error.
A = rand(N,R);
B = rand(M,R);
T = rand(K,R);

Ux = cell(3,1);
Ux{1} = A;
Ux{2} = B;
Ux{3} = T;

Uy = cell(3,1);
Uy{1} = A;
Uy{2} = A;
Uy{3} = T;

rError = fError(X, Y, Ux, Uy, alpha, normX, normY);

fprintf('Non-negative Symmetric Coupled Tensor Factorization\n');
fprintf('Alpha = %f\n', alpha);
fprintf('- Initial error: %f\n', rError);

%% Main Loop: Iterate until convergence
for iter = 1:maxiters
    errorold = rError;
    
    %% Update A
    % the numerator is a sum of matrix times khatri-rao products
    Aupper = mttkrp(X,Ux,1) + alpha*mttkrp(Y, Uy, 2);
    % the denominator involves multiple a khatri-rao products times
    % their transpose.
    AlowerX = (B' * B) .* (T' * T);
    AlowerY = (A' * A) .* (T' * T);
    Alower = A * AlowerX + alpha * A * AlowerY;
    Ares = (Aupper ./ Alower) .^ (1/3);
    Ares(isnan(Ares)) = 1;
    A = A .* Ares;
    
    Ux{1} = A; Uy{1} = A; Uy{2} = A;

    %% Update B
    Bupper = mttkrp(X, Ux, 2);
    BlowerX = (A'*A) .* (T' * T);
    Blower = B * BlowerX;
    Bres = (Bupper ./ Blower);
    Bres(isnan(Bres)) = 1;
    B = B .* Bres;
    Ux{2} = B;

    %% Update T
    Tupper = mttkrp(X, Ux, 3) + alpha*mttkrp(Y, Uy, 3);

    TlowerX = (A' * A) .* (B' * B);
    TlowerY = (A' * A) .* (A' * A);
    Tlower = T * TlowerX + alpha * T * TlowerY;
    Tres = (Tupper ./ Tlower);
    Tres(isnan(Tres)) = 1;
    T = T .* Tres;
    Ux{3} = T; Uy{3} = T;

    rError = fError(X, Y, Ux, Uy, alpha, normX, normY);
    fprintf('- Iteration %d error: %f (error change: %f)\n', ...
            iter, rError, errorold - rError);
end

return;

function e = fError(X, Y, Ux, Uy, alpha, normX, normY)
Xr = ktensor(Ux);
Yr = ktensor(Uy);
e = normX^2 + norm(Xr)^2 - 2 * innerprod(X,Xr) + ...
    alpha * (normY^2 + norm(Yr)^2 - 2 * innerprod(Y,Yr));
return;