package com.maraujo

import java.io.{PrintWriter, BufferedWriter, FileWriter}

import org.rogach.scallop._
import scala.io.Source

import scala.collection.mutable.ArrayBuffer

object ForecastEvalMain {
  class ForecastEvalConf(arguments: Seq[String]) extends ScallopConf(arguments) {
    banner("""
             |Usage: forecastEval <tensor> <A> <B> <T> <rank> <K>
             |
             |  Options:""".stripMargin)

    val tensor = trailArg[String](descr="file with space separated list of non-zeros, one per line.", required = true)

    val A = trailArg[String](descr="file representing the A matrix, <rank> space-separated doubles per line, one line user.", required = true)

    val B = trailArg[String](descr="file representing the B matrix, <rank> space-separated doubles per line, one line community.", required = true)

    val T = trailArg[String](descr="file representing the T matrix, <rank> space-separated doubles per line, one line per timestep.", required = true)

    val rank = trailArg[Int](descr="rank of the decomposition.", required = true)

    val K = trailArg[Int](descr="number of non-zeros to evaluate.", required = true)

    val precisionOut = opt[String](descr = "filename for precision curve.")

    val newPrecisionOut = opt[String](descr = "filename for precision curve of new 'user x community' pairs.")

    val totalPositives = opt[Int](descr = "number of real positives (for recall calculation).")

    val padding = opt[Int](descr = "number of elements to subtract from the time dimension for precision calculations.", default = Some(0))

    val initialT = opt[Int](descr="first timestep of the prediction (0-indexed).", required = true, default = Some(0))

    val forecastPerStep = opt[Boolean](descr = "forecasts K elements per future time-step.", default = Some(false))

    verify()
  }

  def main(args: Array[String]) {
    val opts = new ForecastEvalConf(args)

    println("-- Reading factors")
    // read factors
    println("--- A")
    val Afactors = Array.fill(opts.rank())(ArrayBuffer[Double]())
    Source.fromFile(opts.A()).getLines().foreach(_.split(' ').map(_.toDouble).zipWithIndex.foreach { case (v, k) => Afactors(k).append(v) })
    println("--- B")
    val Bfactors = Array.fill(opts.rank())(ArrayBuffer[Double]())
    Source.fromFile(opts.B()).getLines().foreach(_.split(' ').map(_.toDouble).zipWithIndex.foreach { case (v, k) => Bfactors(k).append(v) })
    println("--- T")
    val Tfactors = Array.fill(opts.rank())(ArrayBuffer[Double]())
    Source.fromFile(opts.T()).getLines().foreach(_.split(' ').map(_.toDouble).zipWithIndex.foreach { case (v, k) => Tfactors(k).append(v) })

    println("-- Creating Forecaster (sorting factor vectors..)")
    // create forecaster
    val forecaster = new ForecastEval(Afactors, Bfactors, Tfactors, opts.rank(), opts.initialT())

    println("-- Reading the tensor")
    // read tensor for evaluation
    val tensor = Source.fromFile(opts.tensor()).getLines().map { s =>
      val Array(x, y, z) = s.split(' ').map(_.toInt)
      (y, x, z)   // note: x and y are swapped because the original file has data in "community x member x timestep" format
    }.toSet

    if (opts.forecastPerStep()) {
      println("-- Forecasting " + opts.K() + " non-zeros per step")
      val forecast = forecaster.nonZerosPerStep(opts.K())
      forecast.zipWithIndex.foreach { case (elems, s) =>
        val tp = elems.count { case (_, elem) =>
          tensor.contains((elem._1, elem._2, elem._3+opts.padding()))
        }
        println("Step " + (s+opts.padding()+opts.initialT()) + " precision " + tp.toDouble/elems.length)
      }
    }
    else {
      println("-- Forecasting top " + opts.K() + " non-zeros.")
      val forecast = forecaster.topNonZeros(opts.K())

      forecast.foreach(x => println(x._2._1 + " " + x._2._2 + " " + x._2._3))

      opts.precisionOut.foreach {
        writeToFile(_, precisionStream(tensor, forecast, opts.totalPositives(), opts.padding()))
      }
      opts.newPrecisionOut.foreach { f =>
        val oldTensorProjected = tensor.filter(p => p._3 < opts.padding()+opts.initialT()).map(p => (p._1, p._2))
        writeToFile(f, newPrecisionStream(tensor, oldTensorProjected, forecast, opts.totalPositives(), opts.padding()))
      }
    }
  }

  def writeToFile(filename: String, stream: Stream[String]) : Unit = {
    val fout = new PrintWriter(new BufferedWriter(new FileWriter(filename)))
    stream.foreach(fout.println)
    fout.close()
  }

  def precisionStream(tensor: Set[(Int, Int, Int)], forecast: Seq[(Double, (Int, Int, Int))], totalPositives: Int, padding: Int): Stream[String] = {
    def next(tp: Int, total: Int) : Stream[String] = {
      if (total == forecast.size) Stream.empty
      else {
        val nz = forecast(total)._2
        val newtp = if (tensor.contains((nz._1, nz._2, nz._3+padding))) tp + 1 else tp
        ((newtp.toDouble / (total + 1)).toString+" "+(newtp.toDouble/totalPositives).toString) #:: next(newtp, total + 1)
      }
    }
    next(0, 0)
  }

  def newPrecisionStream(tensor: Set[(Int, Int, Int)], oldTensorProjected: Set[(Int, Int)], forecast: Seq[(Double, (Int, Int, Int))], totalPositives: Int, padding: Int): Stream[String] = {
    def next(tp: Int, total: Int, nextNz: Int) : Stream[String] = {
      var newNextNz = nextNz
      while (newNextNz < forecast.size && oldTensorProjected.contains((forecast(newNextNz)._2._1, forecast(newNextNz)._2._2)))
        newNextNz += 1
      if (newNextNz == forecast.size)
        Stream.empty
      else {
        val nz = forecast(newNextNz)._2
        println(tp + " " + total + " " + nz)
        val newtp = if (tensor.contains((nz._1, nz._2, nz._3+padding))) tp + 1 else tp
        ((newtp.toDouble / (total + 1)).toString+" "+(newtp.toDouble/totalPositives).toString) #:: next(newtp, total + 1, newNextNz+1)
      }
    }
    next(0, 0, 0)
  }
}